#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
#
#  Copyright 2002  The FreeRADIUS server project
#  Copyright 2002  Boian Jordanov <bjordanov@orbitel.bg>
#  Copyright 2012  Thomas Glanzmann <thomas@glanzmann.de>

use strict;
use Authen::Krb5::Simple;
use BerkeleyDB;
use Mail::Mailer;
use Storable qw(freeze thaw);

use vars qw(%RAD_REQUEST %RAD_REPLY %RAD_CHECK);

use constant RLM_MODULE_REJECT => 0;   # immediately reject the request
use constant RLM_MODULE_FAIL => 1;     # module failed, don't reply
use constant RLM_MODULE_OK => 2;       # the module is OK, continue
use constant RLM_MODULE_HANDLED => 3;  # the module handled the request, so stop.
use constant RLM_MODULE_INVALID => 4;  # the module considers the request invalid.
use constant RLM_MODULE_USERLOCK => 5; # reject the request (user is locked out)
use constant RLM_MODULE_NOTFOUND => 6; # user not found
use constant RLM_MODULE_NOOP => 7;     # module succeeded without doing anything
use constant RLM_MODULE_UPDATED => 8;  # OK (pairs modified)
use constant RLM_MODULE_NUMCODES => 9; # How many return codes there are

my %fail;
my %locked;
my %token;
my %users;
my %ninetofive;
my $otp_lifetime = 600;
my $sender = 'thomas@glanzmann.de';
my $url = 'https://terminal.gmvl.de/';
my $strikes = 3;
my @ninetofive_hours = (7 .. 19);
my @ninetofive_days = (1 .. 5);

my $env = new BerkeleyDB::Env
        -Home   => '/etc/raddb/state',
        -Flags  => DB_CREATE| DB_INIT_CDB | DB_INIT_MPOOL
or die "cannot open environment: $BerkeleyDB::Error";

tie %users, "BerkeleyDB::Hash",
        -Filename       => 'users.db',
        -Flags          => DB_CREATE,
        -Env            => $env
or die "cannot open database: $BerkeleyDB::Error";

my $failed_handle = tie %fail, "BerkeleyDB::Hash",
        -Filename       => 'fail.db',
        -Flags          => DB_CREATE,
        -Env            => $env
or die "cannot open database: $BerkeleyDB::Error";

tie %locked, "BerkeleyDB::Hash",
        -Filename       => 'locked.db',
        -Flags          => DB_CREATE,
        -Env            => $env
or die "cannot open database: $BerkeleyDB::Error";

tie %token, "BerkeleyDB::Hash",
        -Filename       => 'id.db',
        -Flags          => DB_CREATE,
        -Env            => $env
or die "cannot open database: $BerkeleyDB::Error";

tie %ninetofive, "BerkeleyDB::Hash",
        -Filename       => 'ninetofive.db',
        -Flags          => DB_CREATE,
        -Env            => $env
or die "cannot open database: $BerkeleyDB::Error";

sub
send_otp
{
        my %args = @_;

        my $mailer = Mail::Mailer->new('sendmail', '-f', $sender);
        $mailer->open({
                        From => $sender,
                        To => "$args{to}\@sms.glanzmann.de",
                        Subject => $args{otp},
                        });
        $mailer->close();
}

sub
send_email
{
        my %args = @_;

        my $mailer = Mail::Mailer->new('sendmail', '-f', $sender);
        $mailer->open({
                        From => $sender,
                        To => $sender,
                        Subject => "SMSOTP: Account $args{user} locked"
                        });
        print $mailer <<EOF;
Hello,
the user account $args{user} was locked after $strikes failed logins.
In order to unlock the account go to $url
and login as user '$args{user}' with the unlock password '$args{password}'.

Cheers,
        rlm_perl
EOF
        $mailer->close();
}

sub
generate_otp
{
        my $user = lc($RAD_REQUEST{'User-Name'});

        my $mobile;

        if (! defined($mobile = $users{$user})) {
                return RLM_MODULE_REJECT;
        }

        my %token_hash;
        $token_hash{id} = int(1 + rand(99999999));
        $token_hash{otp} = sprintf("%05d", int(1 + rand(99999)));
        $token_hash{time} = time;

        $token{$user} = freeze \%token_hash;

        send_otp(to => $mobile, otp => $token_hash{otp});

        $RAD_REPLY{'Reply-Message'} = 'Enter SMS one time password';
        $RAD_REPLY{'State'} = $token_hash{id};
        $RAD_CHECK{'Response-Packet-Type'} = 'Access-Challenge'; 

        return RLM_MODULE_HANDLED;
}

sub
check_otp
{
        my $user = lc($RAD_REQUEST{'User-Name'});

        my $state = $RAD_REQUEST{'State'};
        $state =~ s/^0x//;
        $state =~ s/([a-fA-F0-9][a-fA-F0-9])/chr(hex($1))/eg;

        my $serialized;
        if (! defined($serialized = $token{$user})) {
                return RLM_MODULE_REJECT;
        }

        my %token_hash = %{ thaw($serialized) };

        if (($token_hash{otp} eq $RAD_REQUEST{'User-Password'})
        && ($token_hash{id} eq $state)
        && ((time() - $token_hash{time}) < $otp_lifetime)) {
                delete($token{$user});
                return RLM_MODULE_OK;
        }

        delete($token{$user});
        return RLM_MODULE_REJECT;
}

sub
generate_password
{
        my $length = 8;
        my $possible = '!"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~';

        my $password;

        my $length_possible = length($possible);

        while (length($password) < $length) {
                $password .= substr($possible, int(rand($length_possible)), 1);
        }

        return $password;
} 

sub
authorize
{
        return RLM_MODULE_OK;
}

sub
user_allowed_to_work_now
{
        my $user = shift || die;

        if (! defined($ninetofive{$user})) {
                return 1;
        }
        my ($sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst) = localtime();
        $year += 1900;
        $month += 1;

        if (! grep $_ == $wday, @ninetofive_days) {
                return 0;
        }

        if (! grep $_ == $hour, @ninetofive_hours) {
                return 0;
        }

        return 1;
}

sub
authenticate
{
        my $user = lc($RAD_REQUEST{'User-Name'});

        if (! defined($users{$user})) {
                return RLM_MODULE_REJECT;
        }

        if (! user_allowed_to_work_now($user)) {
                return RLM_MODULE_REJECT;
        }

        if (defined($locked{$user})) {
                if ($RAD_REQUEST{'User-Password'} eq $locked{$user}) {
                        delete($locked{$user});
                        return RLM_MODULE_REJECT;
                }
                return RLM_MODULE_USERLOCK;
        }

        if (defined($RAD_REQUEST{'State'})) {
                return check_otp();
        }

        if (Authen::Krb5::Simple->new()->authenticate($RAD_REQUEST{'User-Name'}, $RAD_REQUEST{'User-Password'})) {
                delete($fail{$user});
                if (defined($RAD_REQUEST{'Pool-Name'})
                 && $RAD_REQUEST{'Pool-Name'} eq 'smartcard') {
                        return RLM_MODULE_OK;
                }
                return generate_otp();

        } else {
                my $lk = $failed_handle->cds_lock(); 
                my $fails = $fail{$user}++;
                $lk->cds_unlock();
                if ($strikes <= $fails) {
                        delete($fail{$user});
                        delete($token{$user});
                        $locked{$user} = generate_password();
                        send_email(user => $user, password => $locked{$user});
                }
                return RLM_MODULE_REJECT;
        }
}

sub
preacct
{
        return RLM_MODULE_OK;
}

sub
accounting
{
        return RLM_MODULE_OK;
}

sub
checksimul
{
        return RLM_MODULE_OK;
}

sub
pre_proxy
{
        return RLM_MODULE_OK;
}

sub
post_proxy
{
        return RLM_MODULE_OK;
}

sub
post_auth
{
        return RLM_MODULE_OK;
}

sub
xlat
{
        # nothing todo
}

sub
detach
{
        # nothing todo
}
